_restic_backup = """
restic_backup_files_new{{config="{config_name}",repository="{repository}"}} {restic_backup[files][new]}
restic_backup_files_changed{{config="{config_name}",repository="{repository}"}} {restic_backup[files][changed]}
restic_backup_files_unmodified{{config="{config_name}",repository="{repository}"}} {restic_backup[files][unmodified]}
restic_backup_dirs_new{{config="{config_name}",repository="{repository}"}} {restic_backup[dirs][new]}
restic_backup_dirs_changed{{config="{config_name}",repository="{repository}"}} {restic_backup[dirs][changed]}
restic_backup_dirs_unmodified{{config="{config_name}",repository="{repository}"}} {restic_backup[dirs][unmodified]}
restic_backup_processed_files{{config="{config_name}",repository="{repository}"}} {restic_backup[processed][files]}
restic_backup_processed_size_bytes{{config="{config_name}",repository="{repository}"}} {restic_backup[processed][size_bytes]}
restic_backup_processed_duration_seconds{{config="{config_name}",repository="{repository}"}} {restic_backup[processed][duration_seconds]}
restic_backup_added_to_repo{{config="{config_name}",repository="{repository}"}} {restic_backup[added_to_repo]}
restic_backup_rc{{config="{config_name}",repository="{repository}"}} {restic_backup[rc]}
"""

_restic_forget = """
restic_forget_removed_snapshots{{config="{config_name}",repository="{repository}"}} {restic_forget[removed_snapshots]}
restic_forget_rc{{config="{config_name}",repository="{repository}"}} {restic_forget[rc]}
"""

_restic_prune = """
restic_prune_containing_packs_before{{config="{config_name}",repository="{repository}"}} {restic_prune[containing_packs_before]}
restic_prune_containing_blobs{{config="{config_name}",repository="{repository}"}} {restic_prune[containing_blobs]}
restic_prune_containing_size_bytes{{config="{config_name}",repository="{repository}"}} {restic_prune[containing_size_bytes]}
restic_prune_duplicate_blobs{{config="{config_name}",repository="{repository}"}} {restic_prune[duplicate_blobs]}
restic_prune_duplicate_size_bytes{{config="{config_name}",repository="{repository}"}} {restic_prune[duplicate_size_bytes]}
restic_prune_in_use_blobs{{config="{config_name}",repository="{repository}"}} {restic_prune[in_use_blobs]}
restic_prune_removed_blobs{{config="{config_name}",repository="{repository}"}} {restic_prune[removed_blobs]}
restic_prune_invalid_files{{config="{config_name}",repository="{repository}"}} {restic_prune[invalid_files]}
restic_prune_deleted_packs{{config="{config_name}",repository="{repository}"}} {restic_prune[deleted_packs]}
restic_prune_rewritten_packs{{config="{config_name}",repository="{repository}"}} {restic_prune[rewritten_packs]}
restic_prune_size_freed_bytes{{config="{config_name}",repository="{repository}"}} {restic_prune[size_freed_bytes]}
restic_prune_removed_index_files{{config="{config_name}",repository="{repository}"}} {restic_prune[removed_index_files]}
restic_prune_rc{{config="{config_name}",repository="{repository}"}} {restic_prune[rc]}
"""

_restic_check = """
restic_check_errors{{config="{config_name}",repository="{repository}"}} {restic_check[errors]}
restic_check_errors_data{{config="{config_name}",repository="{repository}"}} {restic_check[errors_data]}
restic_check_errors_snapshots{{config="{config_name}",repository="{repository}"}} {restic_check[errors_snapshots]}
restic_check_read_data{{config="{config_name}",repository="{repository}"}} {restic_check[read_data]}
restic_check_check_unused{{config="{config_name}",repository="{repository}"}} {restic_check[check_unused]}
restic_check_rc{{config="{config_name}",repository="{repository}"}} {restic_check[rc]}
"""


def prometheus_generate_lines(metrics, repository, config_name):
    output = 'restic_last_run{{config="{config_name}",repository="{repository}"}} {last_run}\n'
    if metrics.get('restic_backup'):
        output += _restic_backup
    if metrics.get('restic_forget'):
        output += _restic_forget
    if metrics.get('restic_prune'):
        output += _restic_prune
    if metrics.get('restic_check'):
        output += _restic_check
    output = "\n\n"
    return output.format(**metrics, repository=repository, config_name=config_name)


def prometheus_write_file(lines, path):
    with open(path, 'w') as file:
        file.writelines(lines)



from runrestic.metrics.prometheus import prometheus_write_file
from .prometheus import prometheus_generate_lines


def generate_lines(metrics: dict, repository, config_name: str, configuration: dict):
    if 'prometheus' in configuration.keys():
        return prometheus_generate_lines(metrics, repository, config_name)


def write_lines(lines: str, configuration: dict):
    if 'prometheus' in configuration.keys():
        return prometheus_write_file(lines, configuration.get('prometheus')['path'])



import re

re_bytes = re.compile('([0-9.]+) ?([a-zA-Z]*B)')
re_time = re.compile('(?:([0-9]+):)?([0-9]+):([0-9]+)')


def _parse_size(size: str):
    number, unit = re_bytes.findall(size)[0]
    units = {
        "B": 1, "kB": 10 ** 3, "MB": 10 ** 6, "GB": 10 ** 9, "TB": 10 ** 12,
        "KiB": 1024, "MiB": 2 ** 20, "GiB": 2 ** 30, "TiB": 2 ** 40,
    }
    return float(number) * units[unit]


def _parse_time(time: str):
    hours, minutes, seconds = (int(x) if x else 0 for x in re_time.findall(time)[0])
    if minutes:
        seconds += minutes * 60
    if hours:
        seconds += hours * 3600
    return seconds


def parse_backup(output: str) -> dict:
    files_new, files_changed, files_unmodified = re.findall('Files:\s+([0-9]+) new,\s+([0-9]+) changed,\s+([0-9]+) unmodified', output)[0]
    dirs_new, dirs_changed, dirs_unmodified = re.findall('Dirs:\s+([0-9]+) new,\s+([0-9]+) changed,\s+([0-9]+) unmodified', output)[0]
    added_to_the_repo = re.findall('Added to the repo:\s+(-?[0-9.]+ [a-zA-Z]*B)', output)[0]
    processed_files, processed_size, processed_time = re.findall('processed ([0-9]+) files,\s+(-?[0-9.]+ [a-zA-Z]*B) in ([0-9]+:+[0-9]+)', output)[0]

    return {
        'files': {'new': files_new, 'changed': files_changed, 'unmodified': files_unmodified},
        'dirs': {'new': dirs_new, 'changed': dirs_changed, 'unmodified': dirs_unmodified},
        'processed': {'files': processed_files, 'size_bytes': _parse_size(processed_size), 'duration_seconds': _parse_time(processed_time)},
        'added_to_repo': _parse_size(added_to_the_repo)
    }


def parse_forget(output: str) -> dict:
    re_removed_snapshots = re.findall("remove ([0-9]+) snapshots", output)
    return {
        'removed_snapshots': re_removed_snapshots[0] if re_removed_snapshots else 0
    }


def parse_prune(output: str) -> dict:
    containing_packs_before, containing_blobs_before, containing_size_before = \
        re.findall('repository contains ([0-9]+) packs \(([0-9]+) blobs\) with (-?[0-9.]+ ?[a-zA-Z]*B)', output)[0]
    duplicate_blobs, duplicate_size = re.findall('([0-9]+) duplicate blobs, (-?[0-9.]+ ?[a-zA-Z]*B) duplicate', output)[0]
    in_use_blobs, _, removed_blobs = re.findall('found ([0-9]+) of ([0-9]+) data blobs still in use, removing ([0-9]+) blobs', output)[0]
    invalid_files = re.findall('will remove ([0-9]+) invalid files', output)[0]
    deleted_packs, rewritten_packs, size_freed = \
        re.findall('will delete ([0-9]+) packs and rewrite ([0-9]+) packs, this frees (-?[0-9.]+ ?[a-zA-Z]*B)', output)[0]
    removed_index_files = re.findall('remove ([0-9]+) old index files', output)[0]

    return {
        'containing_packs_before': containing_packs_before,
        'containing_blobs': containing_blobs_before,
        'containing_size_bytes': _parse_size(containing_size_before),
        'duplicate_blobs': duplicate_blobs,
        'duplicate_size_bytes': _parse_size(duplicate_size),
        'in_use_blobs': in_use_blobs,
        'removed_blobs': removed_blobs,
        'invalid_files': invalid_files,
        'deleted_packs': deleted_packs,
        'rewritten_packs': rewritten_packs,
        'size_freed_bytes': _parse_size(size_freed),
        'removed_index_files': removed_index_files,
    }



import logging
import os
import sys
from argparse import ArgumentParser

import toml

from runrestic.config import collect, signals, log
from runrestic.config.environment import initialize_environment
from runrestic.metrics import generate_lines, write_lines
from runrestic.restic import ResticRepository

logger = logging.getLogger(__name__)


def parse_arguments():
    parser = ArgumentParser(
        prog='runrestic',
        description='''
            A wrapper for restic. It runs restic based on config files and also outputs metrics.
            If none of the --prune, --create, or --check options are given, then runrestic defaults
            to all three: prune, create, and check archives.
            '''
    )
    parser.add_argument('action', type=str, nargs='*',
                        help='one or more from the following actions: [init,backup,prune,check]')
    parser.add_argument('-n', '--dry-run', dest='dry_run', action='store_true',
                        help='Apply --dry-run where applicable (i.e.: forget)')
    args = parser.parse_args()
    return args


def run_configuration(config_filename, args):
    """
    Parse a single configuration file, and execute its defined backups, pruning, and/or consistency checks.
    """
    logger.info(f'Parsing configuration file: {config_filename}')
    with open(config_filename) as file:
        try:
            config = toml.load(file)
        except toml.TomlDecodeError as e:
            logger.warning(f"Problem parsing {config_filename}: {e}\n")
            return

    config['args'] = args

    if not args.action:
        args.action = ['backup', 'prune', 'check']

    initialize_environment(config.get('environment'))
    log_metrics = config.get('metrics') and not args.dry_run and not args.action == ['init']
    metrics_lines = ""

    for repository in config.get('repositories'):
        repo = ResticRepository(repository, log_metrics, args.dry_run)

        if 'init' in args.action:
            repo.init()
        if 'backup' in args.action:
            repo.backup(config.get('location'))
        if 'prune' in args.action:
            repo.forget(config.get('retention'))
            repo.prune()
        if 'check' in args.action:
            repo.check(config.get('consistency'))

        if log_metrics:
            config_name = config.get('config_name') or os.path.basename(config_filename)
            metrics_lines += generate_lines(repo.log, repository, config_name, config.get('metrics'))
    if log_metrics:
        write_lines(metrics_lines, config.get('metrics'))


def main():
    args = parse_arguments()
    signals.configure_signals()
    log.configure_logging()

    try:
        config_filenames = tuple(collect.collect_config_filenames())

        if len(config_filenames) == 0:
            raise ValueError('Error: No configuration files found in')

        for config_filename in config_filenames:
            run_configuration(config_filename, args)

    except (ValueError, OSError) as error:
        print(error)
        sys.exit(1)



import os
import signal


def _handle_signal(signal_number, frame):  # pragma: no cover
    """
    Send the signal to all processes in runrestic's process group, which includes child process.
    """
    os.killpg(os.getpgrp(), signal_number)


def configure_signals():  # pragma: no cover
    """
    Configure runrestic's signal handlers to pass relevant signals through to any child processes
    like Restic. Note that SIGINT gets passed through even without these changes.
    """
    for signal_number in (signal.SIGHUP, signal.SIGTERM, signal.SIGUSR1, signal.SIGUSR2):
        signal.signal(signal_number, _handle_signal)



import os


def initialize_environment(environment_config):
    if not (environment_config.get('RESTIC_PASSWORD') or environment_config.get('RESTIC_PASSWORD_FILE')):
        raise Exception("You need to specify either a password or a password_file")

    for key, value in environment_config.items():
        os.environ[key] = value



import os


def get_default_config_paths():
    '''
    Based on the value of the XDG_CONFIG_HOME and HOME environment variables, return a list of
    default configuration paths. This includes both system-wide configuration and configuration in
    the current user's home directory.
    '''
    user_config_directory = os.getenv('XDG_CONFIG_HOME') or os.path.expandvars(os.path.join('$HOME', '.config'))

    return [
        '/etc/runrestic.toml',
        '/etc/runrestic',
        f'{user_config_directory}/runrestic',
    ]


def collect_config_filenames():
    '''
    Given a sequence of config paths, both filenames and directories, resolve that to just an
    iterable of files. Accomplish this by listing any given directories looking for contained config
    files (ending with the ".yaml" extension). This is non-recursive, so any directories within the
    given directories are ignored.

    Return paths even if they don't exist on disk, so the user can find out about missing
    configuration paths. However, skip a default config path if it's missing, so the user doesn't
    have to create a default config path unless they need it.
    '''
    real_default_config_paths = set(map(os.path.realpath, get_default_config_paths()))

    for path in real_default_config_paths:
        exists = os.path.exists(path)

        if os.path.realpath(path) in real_default_config_paths and not exists:
            continue

        if not os.path.isdir(path) or not exists:
            yield path
            continue

        for filename in os.listdir(path):
            full_filename = os.path.join(path, filename)
            if full_filename.endswith('.toml') and not os.path.isdir(full_filename):
                yield full_filename



from setuptools import setup, find_packages


VERSION = '0.1.1'

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='runrestic',
    version=VERSION,
    description='A wrapper script for Restic backup software that creates and prunes backups',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Andreas Nüßlein',
    author_email='nuts@noova.de',
    url='https://github.com/andreasnuesslein/runrestic',
    classifiers=(
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python',
        'Topic :: Security :: Cryptography',
        'Topic :: System :: Archiving :: Backup',
    ),
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'runrestic = runrestic.commands.runrestic:main',
        ]
    },
    install_requires=(
        'toml>=0.10.0',
        'setuptools',
    ),
    include_package_data=True,
)



import logging


def configure_logging():
    log = logging.getLogger('runrestic')
    log.setLevel(logging.DEBUG)
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(message)s')
    handler.setFormatter(formatter)
    log.addHandler(handler)



import json
import logging
import subprocess
from datetime import datetime

from runrestic.restic.output_parser import parse_prune, parse_backup, parse_forget

logger = logging.getLogger(__name__)


class ResticRepository:
    def __init__(self, repository, log_metrics, dry_run=False):
        self.basecommand = ['restic', '-r', repository]
        self.log_metrics = log_metrics
        if self.log_metrics:
            self.log = {
                'last_run': datetime.now().timestamp(),
            }
        self.dry_run = dry_run
        self.initialized = self.check_initialization()

    def init(self):
        logger.info('Restic::Init')

        if self.initialized is True:
            logger.warning('Repo already initialized')
            return

        cmd = self.basecommand + ['init']
        output = subprocess.check_output(cmd, stderr=subprocess.STDOUT, universal_newlines=True)
        logger.debug(" ".join(cmd))
        logger.debug(output)

    def snapshots(self):
        cmd = self.basecommand + ['snapshots', '--json']

        try:
            output = subprocess.check_output(cmd, stderr=subprocess.STDOUT, universal_newlines=True)
        except subprocess.CalledProcessError as e:
            if 'config: no such file or directory' in e.output:
                return False
            raise e
        try:
            snapshots_json = json.loads(output)
            # logger.debug(snapshots_json)
        except json.JSONDecodeError as e:
            raise e
        return snapshots_json

    def check_initialization(self):
        snapshots = self.snapshots()
        if snapshots == False:
            return False
        return True

    def backup(self, location):
        logger.info('Restic::Backup')
        if not self.initialized:
            logger.error("Repo is not initialized")
            return

        cmd = self.basecommand + ['backup']

        if not location.get('source_directories'):
            raise Exception("You need to specify source_directories.")
        cmd += location.get('source_directories')

        for exclude_pattern in location.get('exclude_patterns', []):
            cmd += ['--exclude', exclude_pattern]
        for exclude_file in location.get('exclude_files', []):
            cmd += ['--exclude-file', exclude_file]

        try:
            output = subprocess.check_output(cmd, stderr=subprocess.STDOUT, universal_newlines=True)
            process_rc = 0
        except subprocess.CalledProcessError as e:
            output = e.output
            process_rc = e.returncode

        logger.debug(" ".join(cmd))
        logger.debug(output)

        if self.log_metrics:
            self.log['restic_backup'] = parse_backup(output)
            self.log['restic_backup']['rc'] = process_rc

    def forget(self, retention):
        logger.info('Restic::Forget')
        if not self.initialized:
            logger.error("Repo is not initialized")
            return

        cmd = self.basecommand + ['forget']

        if self.dry_run:
            cmd += ['--dry-run']

        for key, value in retention.items():
            if key.startswith('keep-'):
                cmd += [f'--{key}', str(value)]

        try:
            output = subprocess.check_output(cmd, stderr=subprocess.STDOUT, universal_newlines=True)
            process_rc = 0
        except subprocess.CalledProcessError as e:
            output = e.output
            process_rc = e.returncode

        logger.debug(" ".join(cmd))
        logger.debug(output)

        if self.log_metrics:
            self.log['restic_forget'] = parse_forget(output)
            self.log['restic_forget']['rc'] = process_rc

    def prune(self):
        logger.info('Restic::Prune')
        if not self.initialized:
            logger.error("Repo is not initialized")
            return

        cmd = self.basecommand + ['prune']

        try:
            output = subprocess.check_output(cmd, stderr=subprocess.STDOUT, universal_newlines=True)
            process_rc = 0
        except subprocess.CalledProcessError as e:
            output = e.output
            process_rc = e.returncode

        logger.debug(" ".join(cmd))
        logger.debug(output)

        if self.log_metrics:
            self.log['restic_prune'] = parse_prune(output)
            self.log['restic_prune']['rc'] = process_rc

    def check(self, consistency):
        logger.info('Restic::Check')
        if not self.initialized:
            logger.error("Repo is not initialized")
            return

        cmd = self.basecommand + ['check']

        metrics = {
            'errors': False, 'errors_data': False, 'errors_snapshots': False,
            'read_data': False, 'check_unused': False,
        }

        if 'check-unused' in consistency.get('checks'):
            cmd += ['--check-unused']
            metrics['check_unused'] = True

        if 'read-data' in consistency.get('checks'):
            cmd += ['--read-data']
            metrics['read_data'] = True

        try:
            output = subprocess.check_output(cmd, stderr=subprocess.STDOUT, universal_newlines=True)
            process_rc = 0
        except subprocess.CalledProcessError as e:
            output = e.output
            process_rc = e.returncode

            metrics['errors'] = True
            if "error: load <snapshot/" in output:
                metrics['errors_snapshots'] = True
            if "Pack ID does not match," in output:
                metrics['errors_data'] = True

        logger.debug(" ".join(cmd))
        logger.debug(output)

        if self.log_metrics:
            self.log['restic_check'] = metrics
            self.log['restic_check']['rc'] = process_rc
